package tsp;

import java.util.Random;

public class Solution implements Comparable<Solution> {
    // path through cities (solution data)

    private int path[];
    // link to environment
    private Environment env;
    // fitness (-1 -> not calculated yet)
    private float fitness = -1f;

    // random solution instance constructor
    public Solution(Environment env) {
        this.env = env;

        randomPath();
    }

    // copy solution instance constructor
    public Solution(Solution orig) {
        this.env = orig.getEnv();

        int nCities = env.getNCities();

        path = new int[nCities + 1];

        for (int i = 0; i < path.length; i++) {
            path[i] = orig.getPathPoint(i);
        }
    }

    // build a random path for this solution instance
    private void randomPath() {
        int nCities = env.getNCities();

        // path is nCities + 1 in length (must return back to start)
        path = new int[nCities + 1];

        // path must start and end on node 0
        path[0] = path[nCities] = 0;

        // first, set path to be an ordered visit of cities
        for (int i = 1; i < nCities; i++) {
            path[i] = i;
        }

        Random randGen = env.getRandGen();

        // for each node in the path (except first and last), randomly swap it with another
        for (int i = 1; i < nCities; i++) {
            int j;
            do {
                j = randGen.nextInt(nCities - 1) + 1;
            } while (j == i);
            int temp = path[i];
            path[i] = path[j];
            path[j] = temp;
        }
    }

    // randomly select two nodes in the path and swap their positions (for mutation)
    public void randSwap() {
        Random randGen = env.getRandGen();

        int i, j;
        do {
            i = randGen.nextInt(path.length - 2) + 1;
            j = randGen.nextInt(path.length - 2) + 1;
        } while (i == j);

        int temp = path[i];
        path[i] = path[j];
        path[j] = temp;
    }

    public Environment getEnv() {
        return env;
    }

    // get the node # at a specific index along the path
    public int getPathPoint(int i) {
        return path[i];
    }

    // get the fitness value for this solution instance
    public float getFitness() {
        // if not calculated (fitness < 0), calculate fitness
        if (fitness < 0) {
            fitness = 0;

            // just sum up the distances between nodes along the path
            for (int i = 0; i < path.length - 1; i++) {
                fitness += env.getDist(path[i], path[i + 1]);
            }
        }

        return fitness;
    }

    // simple hash code (base on hash code of toString())
    @Override
    public int hashCode() {
        return toString().hashCode();
    }

    // simple equals (base on solution path data)
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (obj == this) {
            return true;
        }
        if (obj.getClass() != getClass()) {
            return false;
        }

        Solution s2 = (Solution) obj;

        for (int i = 0; i < path.length; i++) {
            if (path[i] != s2.getPathPoint(i)) {
                return false;
            }
        }

        return true;
    }

    // simple compareTo (compare fitness values)
    @Override
    public int compareTo(Solution sol) {
        return Float.compare(getFitness(), sol.getFitness());
    }

    // string representation: path + fitness
    public String toString() {
        String pathStr = "[";

        for (int i = 0; i < path.length; i++) {
            pathStr += ((i != 0 ? "," : "") + path[i]);
        }

        pathStr += "]"; // + getFitness(); //String.format("%.1f", getFitness());

        return pathStr;
    }
}
