package tsp;

import java.util.Arrays;
import java.util.HashSet;

public class State extends HashSet<Solution> {

    private Environment env;
    // best solution (so far)
    private Solution bestSol = null;

    public State(Environment env) {
        this.env = env;

        // construct an initial population of a size specified in the environment
        int initialPop = env.getInitialPop();

        for (int i = 0; i < initialPop; i++) {
            Solution sol = new Solution(env);

            // if the newly generated individual (solution) is best, store it
            if (bestSol == null || sol.getFitness() < bestSol.getFitness()) {
                bestSol = sol;
            }

            this.add(sol);
        }
    }

    // transition function
    public void Transition() {
        // select either the best individual or a random individual to mutate (base on env var)
        Solution selected = env.getUseBestSol() ? getBestSol() : getRandSol();

        // copy the selected solution
        Solution mutant = new Solution(selected);

        // and mutate the copy               
        mutant.randSwap();

        // if the new (mutant) individual is better than current best, store it
        if (mutant.getFitness() < bestSol.getFitness()) {
            bestSol = mutant;
        }

        // add mutant to population
        this.add(mutant);

        // if using genocide, and population is greater than max pop, do genocide
        if (env.getUseGenocide() && this.size() > env.getMaxPop()) {
            Genocide();
        }
    }

    // perform genocide
    public void Genocide() {
        // copy the population to an array
        Solution[] population = this.toArray(new Solution[0]);

        // sort the array (individuals are compared by fitness values)
        Arrays.sort(population);

        // keep only the N best solutions, where N = genocided population (env var)
        for (int i = env.getGenPop() - 1; i < population.length; i++) {
            this.remove(population[i]);
        }
    }

    // select a random solution
    public Solution getRandSol() {
        // compute random index
        int selectedIndex = env.getRandGen().nextInt(this.size());
        int i = 0;
        Solution selected = null;

        // and iterate until its found (yea, this is nasty... HashSets don't have random access :P)
        for (Solution s : this) {
            if (i == selectedIndex) {
                selected = s;
                break;
            }
            i++;
        }

        return selected;
    }

    // get the best solution (so far)
    public Solution getBestSol() {
        return bestSol;
    }

    // string representation of state (list of individuals in population)
    public String toString() {
        String retStr = "";

        for (Solution sol : this) {
            retStr += sol.toString() + "\n";
        }

        return retStr;
    }
}
